# SPDX-License-Identifier: MIT
"""Race Result USB Decoder interface."""

from __future__ import division

import threading
import Queue
import decimal
import logging
import serial
from time import sleep

from . import (decoder, DECODER_LOG_LEVEL)
from metarace import tod
from metarace import sysconf
from metarace import strops

_log = logging.getLogger(u'metarace.decoder.rru')
_log.setLevel(logging.DEBUG)

_RRU_BAUD = 19200
_RRU_PASSLEN = 12
_RRU_BEACONLEN = 17
_RRU_LOWBATT = 2.1  # Warn if battery voltage is below this many volts
_RRU_REFCHECK = 1800  # check ref after this many timeouts
_RRU_REFTHRESH = 0x2a30000  # 2 x 86400 x 256
_RRU_ENCODING = u'iso8859-1'
_RRU_MARKER = u'_____127'  # trigger marker
_RRU_EOL = u'\n'
# Serial port I/O read timeout in seconds
_RRU_IOTIMEOUT = 1.0
# List of handled responses from decoder
_REPLIES = [
    u'ASCII',
    u'CONFSET',
    u'CONFGET',
    u'INFOGET',
    u'SITESURVEY',
    u'TIMESTAMPGET',
    u'EPOCHREFGET',
    u'EPOCHREFSET',
    u'EPOCHREFADJ1D',
    u'PASSINGGET',
    u'PASSINGINFOGET',
    u'BEACONGET',
    u'PREWARN',
]
# Documented configuration parameter. If default is not None, the
# value will be set in _sane().
_CONFINFO = {
    u'01': {
        u'label': u'Push Pre-Warn',
        u'default': None,
        u'00': u'disabled',
        u'01': u'enabled'
    },
    u'02': {
        u'label': u'Blink/beep on repeated passing',
        u'default': None,
        u'00': u'disabled',
        u'01': u'enabled'
    },
    u'03': {
        u'label': u'Impulse input or beep output',
        u'default': None,
        u'00': u'impulse-in',
        u'01': u'beep-out'
    },
    u'04': {
        u'label': u'Auto-shutdown on power loss',
        # force disabled since USB cable connection often fails
        u'default': u'00',
        u'00': u'disabled',
        u'01': u'enabled'
    },
    u'05': {
        u'label': u'Operation Mode',
        # assume usb-timing is required unless explicity configured otherwise
        u'default': u'06',
        u'05': u'usb-kiosk',
        u'06': u'usb-timing',
        u'07': u'usb-store&copy'
    },
    u'06': {
        u'label': u'Channel ID',
        u'default': None,
        u'00': u'1',
        u'01': u'2',
        u'02': u'3',
        u'03': u'4',
        u'04': u'5',
        u'05': u'6',
        u'06': u'7',
        u'07': u'8'
    },
    u'07': {
        u'label': u'Loop ID',
        u'default': None,
        u'00': u'1',
        u'01': u'2',
        u'02': u'3',
        u'03': u'4',
        u'04': u'5',
        u'05': u'6',
        u'06': u'7',
        u'07': u'8'
    },
    u'08': {
        u'label': u'Loop Power',
        u'default': None
    },
    u'09': {
        u'label': u'Blink dead-time',
        u'default': None
    },
    u'0a': {
        u'label': u'Charging via USB',
        u'default': None,
        u'00': u'disabled',
        u'01': u'enabled'
    },
    u'0b': {
        u'label': u'Use DTR',
        # always use DTR, unless configured otherwise
        u'default': u'01',
        u'00': u'disabled',
        u'01': u'enabled'
    },
    u'0c': {
        u'label': u'Alternate Channel Switching',
        u'default': None,
        u'00': u'disabled',
        u'01': u'automatic',
        u'02': u'force'
    },
    u'0d': {
        u'label': u'Box Mode',
        u'default': None,
        u'31': u'check',
        u'32': u'deep-sleep',
        u'33': u'health-check',
        u'34': u'tracking',
        u'41': u'usb-timing',
    },
    u'a0': {
        u'label': u'Tray Scan Power',
        u'default': None,
    },
    u'a1': {
        u'label': u'Tray Scan interval',
        u'default': None,
    },
    u'a2': {
        u'label': u'Tray Scan ramp up delay',
        u'default': None,
    },
    u'a3': {
        u'label': u'Tray Scan row to column delay',
        u'default': None,
    },
    u'a4': {
        u'label': u'Repeat Row and Column cycle for another N repetition',
        u'default': None,
    },
    u'b1': {
        u'label': u'CheckSum',
        u'default': None,
        u'0': u'disabled',
        u'1': u'enabled',
    },
    u'b2': {
        u'label': u'Push Passings',
        u'default': None,
        u'0': u'disabled',
        u'1': u'enabled',
    }
}
# Labels for decoder information options
_INFOLBLS = {
    u'01': u'Decoder ID',
    u'02': u'Firmware Major Version',
    u'03': u'Hardware Version',
    u'04': u'Box Type',
    u'05': u'Battery Voltage',
    u'07': u'Battery State',
    u'08': u'Battery Level',
    u'09': u'Internal Temperature',
    u'0a': u'Supply Voltage',
    u'0b': u'Loop Status',
    u'0c': u'Firmware Minor Version',
}
_BOX_TYPES = {
    u'0a': u'active-ext',
    u'1e': u'management-box',
    u'28': u'usb-timing-box'
}
_BATTERY_STATES = {
    u'00': u'Fault',
    u'01': u'Charging',
    u'02': u'Reduced Charging',
    u'03': u'Discharging'
}
_LOOP_STATES = {
    u'00': u'OK',
    u'01': u'Fault',
    u'02': u'Limit',
    u'03': u'Overvoltage Error'
}


class rru(decoder):
    """Race Result USB Active Decoder"""

    def __init__(self):
        decoder.__init__(self)
        self._sitenoise = {
            1: 100,
            2: 100,
            3: 100,
            4: 100,
            5: 100,
            6: 100,
            7: 100,
            8: 100
        }
        self._config = {}
        self._boxname = None
        self._rrustamp = None
        self._rruht = None
        self._error = False
        self._io = None
        self._rdbuf = b''
        self._curreply = None  # current multi-line response mode
        self._lastpassing = 0
        self._lastrequest = None
        self._request_pending = False
        self._allowstored = False
        self._autochannelid = None
        self._refcount = 0

    # API overrides
    def status(self):
        """Request status info from decoder."""
        for c in sorted(_CONFINFO):
            self.write(u''.join([u'CONFGET;', c]))
        for c in sorted(_INFOLBLS):
            self.write(u''.join([u'INFOGET;', c]))
        self.write(u'BEACONGET')
        self.write(u'PASSINGINFOGET')
        self.write(u'TIMESTAMPGET')

    def connected(self):
        return self._boxname is not None and self._io is not None

    # Device-specific functions
    def _close(self):
        self._boxname = None
        if self._io is not None:
            _log.debug(u'Close connection')
            cp = self._io
            self._io = None
            cp.close()

    def _port(self, port):
        """Re-establish connection to supplied device port."""
        self._request_pending = False
        self._rrustamp = None
        self._rruht = None
        self._close()
        self._rdbuf = b''
        _log.debug(u'Connecting to %r', port)
        s = serial.Serial(baudrate=_RRU_BAUD,
                          rtscts=False,
                          timeout=_RRU_IOTIMEOUT)
        s.dtr = 0  # This must be set _before_ open()
        s.port = port
        s.open()
        self._io = s
        self._sane()

    def _sane(self, data=None):
        """Load system config and then check decoder is properly configured"""
        setconfs = {}
        if sysconf.has_option(u'rru', u'allowstored'):
            self._allowstored = strops.confopt_bool(
                sysconf.get(u'rru', u'allowstored'))
            _log.debug(u'Allow stored passings: %r', self._allowstored)
        if sysconf.has_option(u'rru', u'decoderconfig'):
            setconfs = sysconf.get(u'rru', u'decoderconfig')
            _log.debug(u'Loaded %r config options from sysconf', len(setconfs))
        self._config = {}
        for opt in _CONFINFO:
            ko = _CONFINFO[opt]
            kn = ko[u'label']
            if kn in setconfs and setconfs[kn] is not None:
                self._config[opt] = setconfs[kn]
                sv = setconfs[kn]
                if sv in ko:
                    sv = ko[sv]
                _log.debug(u'Set %s: %s', kn, sv)
            elif ko[u'default'] is not None:
                self._config[opt] = ko[u'default']

        # if channel id set already by site survey, keep it
        if self._autochannelid is not None:
            self._config[u'06'] = self._autochannelid

        # Set protocol
        self.write(u'ASCII')
        # Fetch decoder ID
        self.write(u'INFOGET;01')
        # Set options, ordering ensures Box mode is set before impulse in
        for opt in [
                u'05', u'01', u'02', u'04', u'07', u'08', u'09', u'0a', u'0b',
                u'0c', u'0d', u'a0', u'a1', u'a2', u'a3', u'a4', u'b1', u'b2',
                u'03'
        ]:
            if opt in self._config and self._config[opt] is not None:
                self.write(u';'.join([u'CONFSET', opt, self._config[opt]]))
        # Check if site survey is required
        if u'06' in self._config and self._config[u'06'] is not None:
            if self._config[u'06'].lower() == u'auto':
                _log.info(u'Performing site survey to set Channel ID')
                self.write(u'SITESURVEY')
            else:
                self.write(u';'.join([u'CONFSET', u'06', self._config[u'06']]))
        # Request current epoch ref setting
        self.write(u'EPOCHREFGET')
        # Request current number of ticks
        self.write(u'TIMESTAMPGET')
        # Request loop ID
        self.write(u'CONFGET;07')
        # Request current memory status
        self.write(u'PASSINGINFOGET')

    def _sync(self, data=None):
        _log.debug(u'Performing blocking DTR sync')
        # for the purpose of sync, the "epoch" is considered to be
        # midnight localtime of the current day
        self._rrustamp = None
        self._rruht = None
        # Determine the 'reference' epoch
        nt = tod.now()
        ntt = nt.truncate(0)
        ntd = (nt - ntt).timeval
        if ntd < 0.1 or ntd > 0.9:
            _log.debug(u'Sleeping 0.3s')
            sleep(0.3)
            nt = tod.now()
        ntt = nt.truncate(0)
        ett = ntt + tod.ONE
        _log.debug(u'Host reference time: %s', ett.rawtime())
        es = u'EPOCHREFSET;{0:08x}'.format(int(ett.timeval))
        self._write(es)
        _log.debug(u'Waiting for top of second')
        acceptval = tod.tod(u'0.001')
        diff = ett - nt
        while diff > acceptval and diff < tod.ONE:
            sleep(0.0005)
            nt = tod.now()
            diff = ett - nt
        _log.debug(u'Set DTR')
        self._io.dtr = 1
        sleep(0.2)
        _log.debug(u'Clear DTR')
        self._io.dtr = 0

    def _start_session(self, data=None):
        """Reset decoder to start a new session"""
        # On USB decoder, session is always active, this
        # method performs a hard reset/clear/sync
        self._clear(data)
        _log.info(u'Start session')

    def _stop_session(self, data=None):
        """Stop the current timing session"""
        # USB decoder will continue collecting passings, but
        # they are not received by handler - the passings can
        # be fetched by restarting connection without reset
        self._rrustamp = None
        _log.info(u'Stop session')

    def _clear(self, data=None):
        _log.debug(u'Performing box reset')
        self._boxname = None
        self._write(u'RESET')
        while True:
            m = self._readline()
            #_log.debug(u'RECV: %r', m)
            if m == u'AUTOBOOT':
                break
        self._rrustamp = None
        self._lastrequest = 0
        self._lastpassing = 0
        # Queue 'sane' config options before sync request
        self._sane()
        self.sync()

    def _write(self, msg):
        if self._io is not None:
            ob = (msg + _RRU_EOL)
            self._io.write(ob.encode(_RRU_ENCODING))
            #_log.debug(u'SEND: %r', ob)

    def _tstotod(self, ts):
        """Convert a race result timestamp to time of day."""
        ret = None
        try:
            ti = int(ts, 16) - self._rrustamp
            tsec = decimal.Decimal(ti // 256) + decimal.Decimal(ti % 256) / 256
            nsec = (self._rruht.timeval + tsec) % 86400
            if nsec < 0:
                _log.debug(u'Negative timestamp: %r', nsec)
                nsec = 86400 + nsec
            ret = tod.tod(nsec)
        except Exception as e:
            _log.error(u'%s converting timeval %r: %s', e.__class__.__name__,
                       ts, e)
        return ret

    def _confmsg(self, cid, val):
        """Handle config response."""
        lbl = cid
        vbl = val
        if cid in _CONFINFO:
            option = _CONFINFO[cid]
            if u'label' in option:
                lbl = option[u'label']
            if cid == u'08':
                vbl = '{0:d}%'.format(int(val, 16))
            elif cid == u'07':
                vbl = '{0:d}'.format(int(val, 16) + 1)
            else:
                if val in option:
                    vbl = option[val]
        if cid in self._config:
            # check decoder has the desired value
            if val != self._config[cid] and self._config[cid] != u'auto':
                if self._curreply == u'CONFSET':
                    _log.error(
                        u'Error setting config %s, desired:%r actual:%r', lbl,
                        self._config[cid], val)
                else:
                    _log.info(u'Updating config %s: %s => %s', lbl, val,
                              self._config[cid])
                    self.write(u';'.join([u'CONFSET', cid, self._config[cid]]))
            else:
                _log.info(u'Config %s: %s', lbl, vbl)
        else:
            if cid in [u'07', u'08']:
                _log.info(u'Config %s: %s', lbl, vbl)
            else:
                _log.debug(u'Config %s: %s', lbl, vbl)

    def _infomsg(self, pid, val):
        """Show and save decoder info message."""
        lbl = pid
        vbl = val
        if pid in _INFOLBLS:
            lbl = _INFOLBLS[pid]
            if pid == u'01':  # ID
                vbl = u'A-{0:d}'.format(int(val, 16))
                self._boxname = vbl
            elif pid == u'02':  # Firmware
                vbl = u'v{0:0.1f}'.format(int(val, 16) / 10)
            elif pid == u'03':  # Hardware
                vbl = u'v{0:0.1f}'.format(int(val, 16) / 10)
            elif pid == u'04':  # Box Type
                if val in _BOX_TYPES:
                    vbl = _BOX_TYPES[val]
            elif pid == u'05':  # Batt Voltage
                vbl = u'{0:0.1f}V'.format(int(val, 16) / 10)
            elif pid == u'07':  # Battery State
                if val in _BATTERY_STATES:
                    vbl = _BATTERY_STATES[val]
            elif pid == u'08':  # Battery Level
                vbl = u'{0:d}%'.format(int(val, 16))
            elif pid == u'09':  # Int Temp
                vbl = u'{0:d}\xb0C'.format(int(val, 16))
            elif pid == u'0a':  # Supply Voltage
                vbl = u'{0:0.1f}V'.format(int(val, 16) / 10)
            elif pid == u'0b':  # Loop Status
                if val in _LOOP_STATES:
                    vbl = _LOOP_STATES[val]
            _log.info(u'Info %s: %s', lbl, vbl)
        else:
            _log.info(u'Info [undocumented] %s: %s', lbl, vbl)

    def _refgetmsg(self, epoch, stime):
        """Collect the epoch ref and system tick message."""
        self._rruht = tod.mkagg(int(epoch, 16))
        self._rrustamp = int(stime, 16)
        _log.debug(u'Reference ticks: %r @ %r', self._rrustamp,
                   self._rruht.rawtime())

    def _timestampchk(self, ticks):
        """Receive the number of ticks on the decoder."""
        tcnt = int(ticks, 16)
        _log.info(u'Box tick count: %r', tcnt)
        if tcnt > _RRU_REFTHRESH:
            _log.info(u'Tick threshold exceeded, adjusting ref')
            self.write(u'EPOCHREFADJ1D')

    def _passinginfomsg(self, mv):
        """Receive info about internal passing memory."""
        if len(mv) == 5:
            pcount = int(mv[0], 16)
            if pcount > 0:
                pfirst = int(mv[1], 16)
                pftime = self._tstotod(mv[2])
                plast = int(mv[3], 16)
                pltime = self._tstotod(mv[4])
                _log.info(u'Info %r Passings, %r@%s - %r@%s', pcount, pfirst,
                          pftime.rawtime(2), plast, pltime.rawtime(2))
                if plast + 1 < self._lastpassing:
                    _log.warning(
                        'Decoder memory/ID mismatch last=%r, req=%r, clear/sync required.',
                        plast, self._lastpassing)
                    self._lastpassing = plast + 1
            else:
                _log.info(u'Info No Passings')
        else:
            _log.debug(u'Non-passinginfo message: %r', mv)

    def _passingmsg(self, mv):
        """Receive a passing from the decoder."""
        if len(mv) == _RRU_PASSLEN:
            # USB decoder doesn't return passing ID, use internal count
            istr = unicode(self._lastpassing)
            tagid = mv[0]  # [TranspCode:string]
            wuc = mv[1]  # [WakeupCounter:4]
            timestr = mv[2]  # [Time:8]
            hits = mv[3]  # [Hits:2]
            rssi = mv[4]  # [RSSI:2]
            battery = mv[5]  # [Battery:2]
            loopid = mv[8]  # [LoopId:1]
            adata = mv[10]  # [InternalActiveData:2]

            # Check values
            if not loopid:
                loopid = u'C1'  # add faked id for passives
            else:
                loopid = strops.id2chan(int(loopid, 16) + 1)
            activestore = False
            if adata:
                aval = int(adata, 16)
                activestore = (int(adata, 16) & 0x40) == 0x40
            if tagid == _RRU_MARKER:
                tagid = u''

            if battery and tagid:
                try:
                    bv = int(battery, 16) / 10
                    if bv < _RRU_LOWBATT:
                        _log.warning(u'Low battery %s: %0.1fV', tagid, bv)
                except Exception as e:
                    _log.debug(u'%s reading battery voltage: %s',
                               e.__class__.__name__, e)

            if hits and rssi and tagid:
                try:
                    hitcount = int(hits, 16)
                    rssival = int(rssi, 16)
                    twofour = -90 + ((rssival & 0x70) >> 2)
                    lstrength = 1 + (rssival & 0x0f)
                    if lstrength < 5 or twofour < -82 or hitcount < 4:
                        _log.warning(
                            u'Poor read %s: Hits:%d RSSI:%ddBm Loop:%ddB',
                            tagid, hitcount, twofour, lstrength)
                except Exception as e:
                    _log.debug(u'%s reading hits/RSSI: %s',
                               e.__class__.__name__, e)

            # emit a decoder log line TBD
            _log.log(DECODER_LOG_LEVEL, u';'.join(mv))

            # accept valid passings and trigger callback
            t = self._tstotod(timestr)
            if t is not None:
                t.index = istr
                t.chan = loopid
                t.refid = tagid
                t.source = self._boxname
                if not activestore or self._allowstored:
                    self._trig(t)
                else:
                    pass
            self._lastpassing += 1
        elif len(mv) == 2:
            resp = int(mv[0], 16)
            rcount = int(mv[1], 16)
            if resp != self._lastrequest:
                _log.error(u'Sequence mismatch request: %r, response: %r',
                           self._lastrequest, resp)
                self._lastpassing = 0
            elif rcount > 0:
                _log.debug(u'Receiving %r passings', rcount)
        else:
            _log.debug(u'Non-passing message: %r', mv)

    def _beaconmsg(self, mv):
        """Receive a beacon from the decoder."""
        if len(mv) == _RRU_BEACONLEN:
            # noise/transponder averages
            chid = int(mv[5], 16) + 1
            chnoise = 10.0 * int(mv[12], 16)
            tlqi = int(mv[13], 16) / 2.56
            trssi = -90 + int(mv[14], 16)
            _log.info(u'Info Ch {0} Noise: {1:0.0f}%'.format(chid, chnoise))
            _log.info(u'Info Avg LQI: {0:0.0f}%'.format(tlqi))
            _log.info(u'Info Avg RSSI: {}dBm'.format(trssi))
        elif len(mv) == 1:
            bcount = int(mv[0], 16)
            _log.debug(u'Receiving %r beacons', bcount)
        else:
            _log.debug(u'Non-beacon message: %r', mv)

    def _idupdate(self, reqid, minid):
        """Handle an empty PASSINGGET response."""
        resp = int(reqid, 16)
        newid = int(minid, 16)
        if resp != self._lastrequest:
            _log.error(
                u'ID mismatch: request=%r, response=%r:%r, reset to first passing',
                self._lastrequest, resp, newid)
            self._lastpassing = newid
        else:
            if resp < newid:
                _log.warning(u'Data loss: %r passings not received',
                             newid - resp)
                self._lastpassing = newid
            # otherwise no missed passings, ignore error

    def _surveymsg(self, chan, noise):
        """Receive a site survey update."""
        channo = int(chan, 16) + 1
        if channo in self._sitenoise:
            self._sitenoise[channo] = 10 * int(noise, 16)
        else:
            _log.debug(u'Unknown channel in site survey: %r', channo)

    def _chansurf(self):
        """Examine survey for a better channel and hop if needed."""
        ch = None
        cv = 55  # don't accept any channel with noise over 50%
        lv = []
        for c in sorted(self._sitenoise, key=strops.rand_key):
            nv = self._sitenoise[c]
            lv.append(u'{}:{:d}%'.format(c, nv))
            if nv < cv:
                ch = c
                cv = nv
        _log.debug(u'Site survey: %s', u' '.join(lv))
        if ch is not None:
            _log.info(u'Selected channel %r (%d%%)', ch, cv)
            sv = u'{0:02x}'.format(ch - 1)
            self._autochannelid = sv
            m = u'CONFSET;06;{}'.format(sv)
            self.write(m)
        else:
            _log.warning(u'Unable to find a suitable channel')

    def _handlereply(self, mv):
        """Process the body of a decoder response."""
        if self._curreply == u'PASSINGGET':
            self._passingmsg(mv)
        elif self._curreply == u'PASSINGINFOGET':
            self._passinginfomsg(mv)
        elif self._curreply == u'PASSINGIDERROR':
            if len(mv) == 2:
                self._idupdate(mv[0], mv[1])
        elif self._curreply == u'INFOGET':
            if len(mv) == 2:
                self._infomsg(mv[0], mv[1])
        elif self._curreply in [
                u'EPOCHREFGET', u'EPOCHREFSET', u'EPOCHREFADJ1D'
        ]:
            if len(mv) == 2:
                self._refgetmsg(mv[0], mv[1])
        elif self._curreply in [u'CONFGET', u'CONFSET']:
            if len(mv) == 2:
                self._confmsg(mv[0], mv[1])
        elif self._curreply == u'TIMESTAMPGET':
            if len(mv) == 1:
                self._timestampchk(mv[0])
        elif self._curreply == u'SITESURVEY':
            if len(mv) == 2:
                self._surveymsg(mv[0], mv[1])
        elif self._curreply == u'BEACONGET':
            self._beaconmsg(mv)
        else:
            _log.debug(u'%r : %r', self._curreply, mv)

    def _procline(self, l):
        """Handle the next line of response from decoder."""
        mv = l.split(u';')
        if len(mv) > 0:
            if mv[0] == u'#P':
                # pushed passing overrides cmd/reply logic
                self._passingmsg(mv[1:])
            elif mv[0] == u'PREWARN':
                self._curreply = mv[0]
            elif mv[0] == u'EOR':
                if self._curreply in [u'PASSINGGET', u'PASSINGIDERROR']:
                    self._request_pending = False
                if self._curreply == u'SITESURVEY':
                    self._chansurf()
                self._curreply = None
            elif mv[0] in _REPLIES:
                if self._curreply is not None:
                    _log.debug(u'Protocol error: %r not terminated',
                               self._curreply)
                self._curreply = mv[0]
                if mv[1] != u'00':
                    if self._curreply == u'PASSINGGET' and mv[1] == u'10':
                        self._curreply = u'PASSINGIDERROR'
                    else:
                        _log.debug(u'%r error: %r', self._curreply, mv[1])
            else:
                if self._curreply is not None:
                    self._handlereply(mv)

    def _readline(self):
        """Read from the decoder until end of line or timeout condition."""
        ret = None
        ch = self._io.read(1)
        while ch != b'':
            if ch == b'\n':
                if len(self._rdbuf) > 0:
                    # linefeed ends the current 'message'
                    ret = self._rdbuf.lstrip(b'\0').decode(_RRU_ENCODING)
                    self._rdbuf = b''
                else:
                    ret = u'EOR'  # Flag end of response
                break
            else:
                self._rdbuf += ch
            ch = self._io.read(1)
        return ret

    def _request_next(self):
        """Queue a passingget request if the reftime is set."""
        if self._rrustamp is not None:
            if not self._request_pending:
                self._request_pending = True
                es = u'PASSINGGET;{0:08x}'.format(self._lastpassing)
                self._lastrequest = self._lastpassing
                self.write(es)
                self._refcount += 1
                if self._refcount > _RRU_REFCHECK:
                    self.write(u'TIMESTAMPGET')
                    self._refcount = 0

    def run(self):
        """Decoder main loop."""
        _log.debug(u'Starting')
        self._running = True
        self._boxname = None
        while self._running:
            try:
                m = None  # next commmand
                if self._io is not None:
                    # Fetch all responses from unit
                    refetch = False
                    while True:
                        l = self._readline()
                        if l is None:
                            # wait for sitesurvey
                            if self._curreply != u'SITESURVEY':
                                # on time out, request passings
                                refetch = True
                                break
                        else:
                            #_log.debug(u'RECV: %r', l)
                            self._procline(l)
                            if self._curreply == u'PREWARN':
                                # Note: this does not work
                                refetch = True
                            if l == u'EOR':
                                break
                    if refetch:
                        self._request_next()
                    m = self._cqueue.get_nowait()
                else:
                    m = self._cqueue.get()
                self._cqueue.task_done()
                self._proccmd(m)
            except Queue.Empty:
                pass
            except serial.SerialException as e:
                self._close()
                _log.error(u'%s: %s', e.__class__.__name__, e)
            except Exception as e:
                _log.critical(u'%s: %s', e.__class__.__name__, e)
                self._running = False
        self.setcb()
        _log.debug(u'Exiting')



# SPDX-License-Identifier: MIT
"""Generic race handler for trackmeet."""

import gtk
import glib
import gobject
import logging
import os

import metarace
from metarace import tod
from metarace import timy
from metarace import riderdb
from metarace import scbwin
from metarace import uiutil
from metarace import strops
from metarace import report
from metarace import jsonconfig

_log = logging.getLogger(u'metarace.race')
_log.setLevel(logging.DEBUG)

# config version string
EVENT_ID = u'race-2.0'

# race model column constants
COL_BIB = 0
COL_FIRSTNAME = 1
COL_LASTNAME = 2
COL_CLUB = 3
COL_INFO = 4
COL_DNF = 5
COL_PLACE = 6

# scb function key mappings
key_startlist = u'F3'  # show starters in table
key_results = u'F4'  # recalc/show result window
key_lapdown = u'F11'  # decrement tv lap counter

# timing function key mappings
key_armstart = u'F5'  # arm for start/200m impulse
key_showtimer = u'F6'  # show timer
key_armfinish = u'F9'  # arm for finish impulse

# extended function key mappings
key_abort = u'F5'  # + ctrl for clear/abort
key_falsestart = u'F6'  # + ctrl for false start


class race(object):
    """Data handling for scratch, handicap, keirin, derby, etc races."""

    def clearplaces(self):
        """Clear places from data model."""
        for r in self.riders:
            r[COL_PLACE] = u''

    def getrider(self, bib):
        """Return temporary reference to model row."""
        ret = None
        for r in self.riders:
            if r[COL_BIB].decode(u'utf-8') == bib:
                ret = r
                break
        return ret

    def getiter(self, bib):
        """Return temporary iterator to model row."""
        i = self.riders.get_iter_first()
        while i is not None:
            if self.riders.get_value(i, COL_BIB).decode(u'utf-8') == bib:
                break
            i = self.riders.iter_next(i)
        return i

    def delayed_reorder(self):
        """Call reorder if the flag is one."""
        if self.reorderflag > 1:
            self.reorderflag -= 1
        elif self.reorderflag == 1:
            self.reorder_handicap()
            self.reorderflag = 0
        else:
            self.reorderflag = 0  # clamp negatives
        return False

    def addrider(self, bib=u'', info=None):
        """Add specified rider to race model."""
        nr = [bib, u'', u'', u'', u'', False, u'']
        er = self.getrider(bib)
        if not bib or er is None:
            dbr = self.meet.rdb.getrider(bib, self.series)
            if dbr is not None:
                for i in range(1, 4):
                    nr[i] = self.meet.rdb.getvalue(dbr, i)  # unicode
                    if self.evtype == u'handicap':  # reqd?
                        if info:
                            nr[COL_INFO] = info
            if self.evtype in [u'handicap', u'keirin'] and not self.onestart:
                self.reorderflag += 1
                glib.timeout_add_seconds(1, self.delayed_reorder)
            self.riders.append(nr)
        else:
            if er is not None:
                # Rider already in the model, set the info if
                # event type is handicap or event is part of omnium
                if self.evtype == u'handicap' or self.inomnium:
                    if not er[COL_INFO] and info:  # don't overwrite if set
                        er[COL_INFO] = info

    def dnfriders(self, biblist=u''):
        """Remove listed bibs from the race."""
        for bib in biblist.split():
            r = self.getrider(bib)
            if r is not None:
                r[COL_DNF] = True
                _log.info(u'Rider %r withdrawn', bib)
            else:
                _log.warn(u'Did not withdraw rider %r', bib)
        return False

    def delrider(self, bib):
        """Remove the specified rider from the model."""
        i = self.getiter(bib)
        if i is not None:
            self.riders.remove(i)

    def placexfer(self, placestr):
        """Transfer places in placestr to model."""
        self.clearplaces()
        self.results = []
        placeset = set()
        resname_w = self.meet.scb.linelen - 11

        # move dnf riders to bottom of list and count inriders
        cnt = 0
        incnt = 0
        reorddnf = []
        if len(self.riders) > 0:
            for r in self.riders:
                if r[COL_DNF]:
                    reorddnf.append(cnt)
                else:
                    incnt += 1
                    reorddnf.insert(0, cnt)
                cnt += 1
            self.riders.reorder(reorddnf)

        # update eliminated rider ranks
        outriders = []
        for bib in self.eliminated:
            r = self.getrider(bib)
            rank = incnt + self.startplace
            r[COL_PLACE] = rank
            club = r[COL_CLUB].decode(u'utf-8')
            if len(club) > 3:
                # look it up?
                #if self.series in self.meet.ridermap:
                #rh = self.meet.ridermap[self.series][bib]
                #if rh is not None:
                #club = rh[u'note']
                club = club[0:3]
            outriders.insert(0, [
                unicode(rank) + u'.', bib,
                strops.fitname(r[COL_FIRSTNAME].decode(u'utf-8'),
                               r[COL_LASTNAME].decode(u'utf-8'), resname_w),
                club
            ])
            i = self.getiter(bib)
            incnt -= 1
            self.riders.swap(self.riders.get_iter(incnt), i)

        # overwrite eliminations from placed riders - but warn on overlap
        place = 1
        count = 0
        clubmode = self.meet.get_clubmode()
        for placegroup in placestr.split():
            for bib in placegroup.split(u'-'):
                if bib not in placeset:
                    if count >= incnt and not clubmode:
                        _log.error(u'Error: More places than available')
                        break
                    placeset.add(bib)
                    r = self.getrider(bib)
                    if r is None:  # ensure rider exists at this point
                        _log.warn(u'Adding non-starter: %r', bib)
                        self.addrider(bib)
                        r = self.getrider(bib)
                    rank = place + self.startplace
                    r[COL_PLACE] = rank
                    club = r[COL_CLUB].decode(u'utf-8')
                    if len(club) > 3:
                        # look it up?
                        #if self.series in self.meet.ridermap:
                        #rh = self.meet.ridermap[self.series][bib]
                        #if rh is not None:
                        #club = rh[u'note']
                        club = club[0:3]
                    self.results.append([
                        unicode(rank) + u'.', bib,
                        strops.fitname(r[COL_FIRSTNAME].decode(u'utf-8'),
                                       r[COL_LASTNAME].decode(u'utf-8'),
                                       resname_w), club
                    ])
                    i = self.getiter(bib)
                    self.riders.swap(self.riders.get_iter(count), i)
                    count += 1
                else:
                    _log.error(u'Ignoring duplicate no: %r', bib)
            place = count + 1
        self.results.extend(outriders)
        if count > 0 or len(outriders) > 0:
            self.onestart = True
        if count == incnt:
            self.resulttype = u'RESULT'
        elif count < incnt and len(outriders) > 0:
            self.resulttype = u'STANDING'
        else:
            self.resulttype = u'PROVISIONAL RESULT'

    def loadconfig(self):
        """Load race config from disk."""
        self.riders.clear()
        # set defaults timetype based on event type
        deftimetype = u'start/finish'
        defdistance = None
        defdistunits = u'laps'
        self.seedsrc = None  # default is no seed info
        if self.evtype == u'handicap':
            self.seedsrc = 3  # fetch handicap info from autospec
        if self.evtype in [u'sprint', u'keirin']:
            deftimetype = u'200m'
            defdistunits = u'metres'
            defdistance = u'200'
        if self.evtype == u'elimination':
            i = self.action_model.append([u'Eliminate', u'out'])
            self.action_model.append([u'Un-Eliminate', u'in'])
            if i is not None:
                self.ctrl_action_combo.set_active_iter(i)
        cr = jsonconfig.config({
            u'event': {
                u'startlist': u'',
                u'id': EVENT_ID,
                u'ctrl_places': u'',
                u'eliminated': [],
                u'start': None,
                u'lstart': None,
                u'comments': [],
                u'finish': None,
                u'runlap': None,
                u'distance': defdistance,
                u'distunits': defdistunits,
                u'showinfo': True,
                u'inomnium': False,
                u'startplace': 0,
                u'autospec': u'',
                u'timetype': deftimetype
            },
            u'riders': {}
        })
        cr.add_section(u'event')
        cr.add_section(u'riders')
        if os.path.exists(self.configfile):
            try:
                with open(self.configfile, 'rb') as f:
                    cr.read(f)
            except Exception as e:
                _log.error(u'Unable to read config: %s', e)
        else:
            _log.info(u'%r not found, loading defaults', self.configfile)
        self.inomnium = strops.confopt_bool(cr.get(u'event', u'inomnium'))
        if self.inomnium:
            self.seedsrc = 1  # fetch start list seeding from omnium
        self.autospec = cr.get(u'event', u'autospec')
        rlist = cr.get(u'event', u'startlist').split()
        for r in rlist:
            nr = [r, u'', u'', u'', u'', False, u'']
            if cr.has_option('riders', r):
                ril = cr.get(u'riders', r)
                for i in range(0, 3):
                    if len(ril) > i:
                        nr[i + 4] = ril[i]
            # Re-patch name
            dbr = self.meet.rdb.getrider(r, self.series)
            if dbr is not None:
                for i in range(1, 4):
                    nr[i] = self.meet.rdb.getvalue(dbr, i)  # unicode
            self.riders.append(nr)

        # race infos
        self.comments = cr.get(u'event', u'comments')
        self.startplace = strops.confopt_posint(cr.get(u'event',
                                                       u'startplace'))
        self.set_timetype(cr.get(u'event', u'timetype'))
        self.distance = strops.confopt_dist(cr.get(u'event', u'distance'))
        self.units = strops.confopt_distunits(cr.get(u'event', u'distunits'))
        self.runlap = strops.confopt_posint(cr.get(u'event', u'runlap'))
        if self.timetype != u'200m' and self.event[u'laps']:
            # use event program to override
            self.units = u'laps'
            self.distance = strops.confopt_posint(self.event[u'laps'],
                                                  self.distance)
        self.update_expander_lbl_cb()
        self.info_expand.set_expanded(
            strops.confopt_bool(cr.get(u'event', u'showinfo')))
        self.set_start(cr.get(u'event', u'start'), cr.get(u'event', u'lstart'))
        self.set_finish(cr.get(u'event', u'finish'))
        self.set_elapsed()
        self.eliminated = cr.get(u'event', u'eliminated')
        places = strops.reformat_placelist(cr.get(u'event', u'ctrl_places'))
        self.ctrl_places.set_text(places)
        self.placexfer(places)
        if places:
            self.doscbplaces = False  # only show places on board if not set
            self.setfinished()
        else:
            if self.autospec:
                self.meet.autostart_riders(self,
                                           self.autospec,
                                           infocol=self.seedsrc)
            if self.evtype in [u'handicap', u'keirin'] or self.inomnium:
                self.reorder_handicap()

        # After load complete - check config and report.
        eid = cr.get(u'event', u'id')
        if eid and eid != EVENT_ID:
            _log.info(u'Event config mismatch: %r != %r', eid, EVENT_ID)

    def sort_riderno(self, x, y):
        """Sort riders by rider no."""
        return cmp(strops.riderno_key(x[1]), strops.riderno_key(y[1]))

    def sort_handicap(self, x, y):
        """Sort function for handicap marks."""
        if x[2] != y[2]:
            if x[2] is None:  # y sorts first
                return 1
            elif y[2] is None:  # x sorts first
                return -1
            else:  # Both should be ints here
                return cmp(x[2], y[2])
        else:  # Defer to rider number
            return cmp(strops.riderno_key(x[1]), strops.riderno_key(y[1]))

    def reorder_handicap(self):
        """Reorder rider model according to the handicap marks."""
        if len(self.riders) > 1:
            auxmap = []
            cnt = 0
            for r in self.riders:
                auxmap.append([
                    cnt, r[COL_BIB].decode(u'utf-8'),
                    strops.mark2int(r[COL_INFO].decode(u'utf-8'))
                ])
                cnt += 1
            if self.inomnium or self.evtype == u'handicap':
                auxmap.sort(self.sort_handicap)
            else:
                auxmap.sort(self.sort_riderno)
            self.riders.reorder([a[0] for a in auxmap])

    def set_timetype(self, data=None):
        """Update state and ui to match timetype."""
        if data is not None:
            self.timetype = strops.confopt_pair(data, u'200m', u'start/finish')
            self.finchan = 1
            if self.timetype == u'200m':
                self.startchan = 4
            else:
                self.startchan = 0

    def set_start(self, start=None, lstart=None):
        """Set the race start."""
        self.start = tod.mktod(start)
        if lstart is not None:
            self.lstart = tod.mktod(lstart)
        else:
            self.lstart = self.start
        if self.start is None:
            pass
        else:
            if self.finish is None:
                self.setrunning()

    def set_finish(self, finish=None):
        """Set the race finish."""
        self.finish = tod.mktod(finish)
        if self.finish is None:
            if self.start is not None:
                self.setrunning()
        else:
            if self.start is None:
                self.set_start(0)  # TODO: Verify this path
            self.setfinished()

    def log_elapsed(self):
        """Log race elapsed time on Timy."""
        self.meet.main_timer.printline(self.meet.racenamecat(self.event))
        self.meet.main_timer.printline(u'      ST: ' + self.start.timestr(4))
        self.meet.main_timer.printline(u'     FIN: ' + self.finish.timestr(4))
        self.meet.main_timer.printline(u'    TIME: ' +
                                       (self.finish - self.start).timestr(2))

    def set_elapsed(self):
        """Update elapsed time in race ui and announcer."""
        if self.start is not None and self.finish is not None:
            et = self.finish - self.start
            self.time_lbl.set_text(et.timestr(2))
        elif self.start is not None:  # Note: uses 'local start' for RT
            runtm = (tod.now() - self.lstart).timestr(1)
            self.time_lbl.set_text(runtm)
            if self.runlap is not None:
                if self.runlap != self.lastrunlap:
                    _log.debug(u'Runlap: %r', self.runlap)
                    self.lastrunlap = self.runlap
        elif self.timerstat == u'armstart':
            self.time_lbl.set_text(u'       0.0   ')  # tod.ZERO.timestr(1)
            if self.runlap and self.runlap != self.lastrunlap:
                _log.debug(u'Runlap: %r', self.runlap)
                self.lastrunlap = self.runlap
        else:
            self.time_lbl.set_text(u'')

    def delayed_announce(self):
        """Initialise the announcer's screen after a delay."""
        if self.winopen:
            # clear page
            self.meet.txt_clear()
            self.meet.txt_title(u' '.join([
                u'Event', self.evno, u':', self.event[u'pref'],
                self.event[u'info']
            ]))
            self.meet.txt_line(1)
            self.meet.txt_line(19)

            # write out riders
            count = 0
            curline = 4
            posoft = 0
            for r in self.riders:
                count += 1
                if count == 14:
                    curline = 4
                    posoft = 41
                xtra = '    '
                if r[COL_INFO]:
                    inf = r[COL_INFO].decode(u'utf-8')
                    if self.evtype in [u'keirin', u'sprint']:  # encirc draw no
                        inf = strops.drawno_encirc(inf)
                    xtra = strops.truncpad(inf, 4, u'r')
                clubstr = u''
                tcs = r[COL_CLUB].decode(u'utf-8')
                if tcs and tcs <= 3:
                    clubstr = u' (' + tcs + u')'
                namestr = strops.truncpad(
                    strops.fitname(r[COL_FIRSTNAME].decode(u'utf-8'),
                                   r[COL_LASTNAME].decode(u'utf-8'),
                                   25 - len(clubstr),
                                   trunc=True) + clubstr,
                    25,
                    ellipsis=False)

                placestr = u'   '
                if r[COL_PLACE] != u'':
                    placestr = strops.truncpad(
                        r[COL_PLACE].decode(u'utf-8') + u'.', 3)
                elif r[COL_DNF]:
                    placestr = u'dnf'
                bibstr = strops.truncpad(r[COL_BIB].decode(u'utf-8'), 3, u'r')
                self.meet.txt_postxt(
                    curline, posoft,
                    u' '.join([placestr, bibstr, namestr, xtra]))
                curline += 1

            tp = u''
            if self.start is not None and self.finish is not None:
                et = self.finish - self.start
                if self.timetype == u'200m':
                    tp = u'200m: '
                else:
                    tp = u'Time: '
                tp += et.timestr(2) + u'    '
                dist = self.meet.get_distance(self.distance, self.units)
                if dist:
                    tp += u'Avg: ' + et.speedstr(dist)
            self.meet.txt_setline(21, tp)
        return False

    def startlist_report(self, program=False):
        """Return a startlist report."""
        ret = []
        sec = None
        etype = self.event[u'type']
        twocol = True
        if not self.inomnium and not program and etype in [u'axe', u'run']:
            sec = report.section()  # one column overrides
        else:
            sec = report.twocol_startlist()
        headvec = []
        if etype != u'break':
            headvec.extend([u'Event', self.evno, u':'])
        headvec.append(self.event[u'pref'])
        headvec.append(self.event[u'info'])
        if not program:
            headvec.append(u'- Start List')
        sec.heading = u' '.join(headvec)
        lapstring = strops.lapstring(self.event[u'laps'])
        substr = u' '.join(
            [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
        if substr:
            sec.subheading = substr

        self.reorder_handicap()
        sec.lines = []
        cnt = 0
        col2 = []
        if self.inomnium and len(self.riders) > 0:
            sec.lines.append([u' ', u' ', u'The Fence', None, None, None])
            col2.append([u' ', u' ', u'Sprinters Lane', None, None, None])
        for r in self.riders:
            cnt += 1
            rno = r[COL_BIB].decode(u'utf-8')
            rh = self.meet.newgetrider(rno, self.series)
            inf = r[COL_INFO].decode(u'utf-8')
            if self.evtype in [u'keirin', u'sprint']:  # encirc draw no
                inf = strops.drawno_encirc(inf)
            if self.inomnium:
                # inf holds seed, ignore for now
                inf = None
            rname = u''
            if rh is not None:
                rname = rh[u'namestr']
            if self.inomnium:
                if cnt % 2 == 1:
                    sec.lines.append([None, rno, rname, inf, None, None])
                else:
                    col2.append([None, rno, rname, inf, None, None])
            else:
                sec.lines.append([None, rno, rname, inf, None, None])
        for i in col2:
            sec.lines.append(i)
        if self.event[u'plac']:
            while cnt < self.event[u'plac']:
                sec.lines.append([None, None, None, None, None, None])
                cnt += 1
        ret.append(sec)

        ptype = u'Riders'
        if etype == u'run':
            ptype = u'Runners'
        elif self.evtype == u'axe':
            ptype = u'Axemen'
        if cnt > 0 and not program:
            sec = report.bullet_text()
            sec.lines.append([None, u'Total ' + ptype + u': ' + str(cnt)])
            ret.append(sec)
        return ret

    def get_startlist(self):
        """Return a list of bibs in the rider model."""
        ret = []
        for r in self.riders:
            ret.append(r[COL_BIB].decode(u'utf-8'))
        return u' '.join(ret)

    def saveconfig(self):
        """Save race to disk."""
        if self.readonly:
            _log.error(u'Attempt to save readonly event')
            return
        cw = jsonconfig.config()
        cw.add_section(u'event')
        if self.start is not None:
            cw.set(u'event', u'start', self.start.rawtime())
        if self.lstart is not None:
            cw.set(u'event', u'lstart', self.lstart.rawtime())
        if self.finish is not None:
            cw.set(u'event', u'finish', self.finish.rawtime())
        cw.set(u'event', u'ctrl_places',
               self.ctrl_places.get_text().decode(u'utf-8'))
        cw.set(u'event', u'eliminated', self.eliminated)
        cw.set(u'event', u'startlist', self.get_startlist())
        cw.set(u'event', u'showinfo', self.info_expand.get_expanded())
        cw.set(u'event', u'distance', self.distance)
        cw.set(u'event', u'distunits', self.units)
        cw.set(u'event', u'timetype', self.timetype)
        if self.runlap is not None:
            cw.set(u'event', u'runlap', self.runlap)
        cw.set(u'event', u'autospec', self.autospec)
        cw.set(u'event', u'inomnium', self.inomnium)
        cw.set(u'event', u'comments', self.comments)
        cw.set(u'event', u'startplace', self.startplace)

        cw.add_section(u'riders')
        for r in self.riders:
            cw.set(u'riders', r[COL_BIB].decode(u'utf-8'), [
                r[COL_INFO].decode(u'utf-8'), r[COL_DNF],
                r[COL_PLACE].decode(u'utf-8')
            ])
        cw.set(u'event', u'id', EVENT_ID)
        _log.debug(u'Saving event config %r', self.configfile)
        with metarace.savefile(self.configfile) as f:
            cw.write(f)

    def shutdown(self, win=None, msg=u'Exit'):
        """Terminate event object."""
        rstr = u''
        if self.readonly:
            rstr = u'readonly '
        _log.debug(u'Shutdown %sevent %s: %s', rstr, self.evno, msg)
        if not self.readonly:
            self.saveconfig()
        self.winopen = False

    def do_properties(self):
        """Run event properties dialog."""
        prfile = os.path.join(metarace.UI_PATH, u'race_properties.ui')
        _log.debug(u'Building event properties from %r', prfile)
        b = gtk.Builder()
        b.add_from_file(prfile)
        dlg = b.get_object(u'properties')
        dlg.set_transient_for(self.meet.window)
        rt = b.get_object(u'race_score_type')
        if self.timetype != u'200m':
            rt.set_active(0)
        else:
            rt.set_active(1)
        di = b.get_object(u'race_dist_entry')
        if self.distance is not None:
            di.set_text(str(self.distance))
        else:
            di.set_text(u'')
        du = b.get_object(u'race_dist_type')
        if self.units == u'metres':
            du.set_active(0)
        else:
            du.set_active(1)
        se = b.get_object(u'race_series_entry')
        se.set_text(self.series)
        as_e = b.get_object(u'auto_starters_entry')
        as_e.set_text(self.autospec)
        response = dlg.run()
        if response == 1:  # id 1 set in glade for "Apply"
            _log.debug(u'Updating event properties')
            if rt.get_active() == 0:
                self.set_timetype(u'start/finish')
            else:
                self.set_timetype(u'200m')
            dval = di.get_text().decode(u'utf-8')
            if dval.isdigit():
                self.distance = int(dval)
            else:
                self.distance = None
            if du.get_active() == 0:
                self.units = u'metres'
            else:
                self.units = u'laps'

            # update series
            ns = se.get_text().decode(u'utf-8')
            if ns != self.series:
                self.series = ns
                self.event[u'seri'] = ns

            # update auto startlist spec
            nspec = as_e.get_text().decode(u'utf-8')
            if nspec != self.autospec:
                self.autospec = nspec
                if not self.ctrl_places.get_text():
                    if self.autospec:
                        self.meet.autostart_riders(self, self.autospec,
                                                   self.seedsrc)
                    if self.evtype == u'handicap':
                        self.reorder_handicap()

            # xfer starters if not empty
            slist = strops.riderlist_split(
                b.get_object(u'race_starters_entry').get_text().decode(
                    u'utf-8'), self.meet.rdb, self.series)
            for s in slist:
                self.addrider(s)
            glib.idle_add(self.delayed_announce)
        else:
            _log.debug(u'Edit event properties cancelled')

        # if prefix is empty, grab input focus
        if not self.prefix_ent.get_text():
            self.prefix_ent.grab_focus()
        dlg.destroy()

    def resettimer(self):
        """Reset race timer."""
        self.finish = None
        self.start = None
        self.lstart = None
        self.timerstat = u'idle'
        self.ctrl_places.set_text(u'')
        self.eliminated = []
        self.placexfer(u'')
        self.meet.main_timer.dearm(self.startchan)
        self.meet.main_timer.dearm(self.finchan)
        self.stat_but.buttonchg(uiutil.bg_none, u'Idle')
        self.stat_but.set_sensitive(True)
        self.set_elapsed()
        _log.info(u'Event reset - all places cleared.')

    def setrunning(self):
        """Set timer state to 'running'."""
        self.timerstat = u'running'
        self.stat_but.buttonchg(uiutil.bg_none, u'Running')

    def setfinished(self):
        """Set timer state to 'finished'."""
        self.timerstat = u'finished'
        self.stat_but.buttonchg(uiutil.bg_none, u'Finished')
        self.stat_but.set_sensitive(False)
        self.ctrl_places.grab_focus()

    def armstart(self):
        """Toggle timer arm start state."""
        if self.timerstat == u'idle':
            self.timerstat = u'armstart'
            self.stat_but.buttonchg(uiutil.bg_armstart, u'Arm Start')
            self.meet.main_timer.arm(self.startchan)
            if self.timetype == u'200m':
                # also accept C0 on sprint types
                self.meet.main_timer.arm(0)
        elif self.timerstat == u'armstart':
            self.timerstat = u'idle'
            self.time_lbl.set_text(u'')
            self.stat_but.buttonchg(uiutil.bg_none, u'Idle')
            self.meet.main_timer.dearm(self.startchan)
            if self.timetype == u'200m':
                # also accept C0 on sprint types
                self.meet.main_timer.dearm(0)

    def armfinish(self):
        """Toggle timer arm finish state."""
        if self.timerstat == u'running':
            self.timerstat = u'armfinish'
            self.stat_but.buttonchg(uiutil.bg_armfin, u'Arm Finish')
            self.meet.main_timer.arm(self.finchan)
        elif self.timerstat == u'armfinish':
            self.timerstat = u'running'
            self.stat_but.buttonchg(uiutil.bg_none, u'Running')
            self.meet.main_timer.dearm(self.finchan)
        return False  # for use in delayed callback

    def showtimer(self):
        """Display the running time on the scoreboard."""
        if self.timerstat == u'idle':
            self.armstart()
        tp = u'Time:'
        if self.timetype == u'200m':
            tp = u'200m:'
        self.meet.cmd_announce(u'eliminated', u'')
        self.meet.scbwin = scbwin.scbtimer(scb=self.meet.scb,
                                           line1=self.meet.racenamecat(
                                               self.event),
                                           line2=u'',
                                           timepfx=tp)
        wastimer = self.timerwin
        self.timerwin = True
        if self.timerstat == u'finished':
            if not wastimer:
                self.meet.scbwin.reset()
            if self.start is not None and self.finish is not None:
                elap = self.finish - self.start
                self.meet.scbwin.settime(elap.timestr(2))
                dist = self.meet.get_distance(self.distance, self.units)
                if dist:
                    self.meet.scbwin.setavg(elap.speedstr(dist))
            self.meet.scbwin.update()
        else:
            self.meet.scbwin.reset()

    def key_event(self, widget, event):
        """Race window key press handler."""
        if event.type == gtk.gdk.KEY_PRESS:
            key = gtk.gdk.keyval_name(event.keyval) or u'None'
            if event.state & gtk.gdk.CONTROL_MASK:
                if key == key_abort:  # override ctrl+f5
                    self.resettimer()
                    return True
            if key[0] == u'F':
                if key == key_armstart:
                    self.armstart()
                    return True
                elif key == key_armfinish:
                    self.armfinish()
                    return True
                elif key == key_showtimer:
                    self.showtimer()
                    return True
                elif key == key_startlist:
                    self.do_startlist()
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_results:
                    self.doscbplaces = True  # override if already clear
                    self.do_places()
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_lapdown:
                    if self.runlap is not None and self.runlap > 0:
                        self.runlap -= 1
                    return True
        return False

    def do_places(self):
        """Update model and show race result on scoreboard."""
        secs = self.result_report()  # NOTE: calls into placexfer
        self.timerwin = False
        tp = u'Time:'
        if self.start is not None and self.finish is None:
            self.finish = tod.now()
            if self.lstart is not None:
                self.start = self.lstart  # override with localtime
            self.set_elapsed()
        if self.timetype == u'200m':
            tp = u'200m:'
            # previously, winner was displayed on gemini here
        ts = None
        if self.start is not None and self.finish is not None:
            ts = (self.finish - self.start).timestr(2)
        if self.doscbplaces:
            FMT = [(3, u'l'), (3, u'r'), u' ',
                   (self.meet.scb.linelen - 11, u'l'), (4, u'r')]
            self.meet.scbwin = scbwin.scbtable(scb=self.meet.scb,
                                               head=self.meet.racenamecat(
                                                   self.event),
                                               subhead=self.resulttype,
                                               coldesc=FMT,
                                               rows=self.results,
                                               timepfx=tp,
                                               timestr=ts)
            self.meet.scbwin.reset()
            self.doscbplaces = False
        self.setfinished()

    def do_startlist(self):
        """Show start list on scoreboard."""

        self.reorder_handicap()
        self.meet.scbwin = None
        self.timerwin = False
        startlist = []
        ##self.meet.announce.gfx_overlay(1)
        ##self.meet.announce.gfx_set_title(u'Startlist: '
        ##+ self.event[u'pref'] + u' '
        ##+ self.event[u'info'])
        name_w = self.meet.scb.linelen - 8
        for r in self.riders:
            if not r[COL_DNF]:
                nfo = r[COL_INFO].decode(u'utf-8')
                if self.evtype in [u'sprint']:  # add asterisk
                    nfo = nfo + r[COL_CLUB].decode(u'utf-8').rjust(3)
                if not nfo:
                    nfo = r[COL_CLUB].decode(u'utf-8')
                    if len(nfo) > 3:
                        # look it up?
                        #if self.series in self.meet.ridermap:
                        #rh = self.meet.ridermap[self.series][r[0]]
                        #if rh is not None:
                        #nfo = rh[u'note']
                        nfo = nfo[0:3]
                startlist.append([
                    r[COL_BIB].decode(u'utf-8'),
                    strops.fitname(r[COL_FIRSTNAME].decode(u'utf-8'),
                                   r[COL_LASTNAME].decode(u'utf-8'), name_w),
                    nfo
                ])
                #inf = r[COL_INFO].decode(u'utf-8').strip()
                #if self.evtype in [u'keirin', u'sprint']:  # encirc draw no
                #inf = strops.drawno_encirc(inf)
                ##self.meet.announce.gfx_add_row([r[COL_BIB].decode(u'utf-8'),
                ##strops.resname(r[COL_FIRSTNAME].decode(u'utf-8'),
                ##r[COL_LASTNAME].decode(u'utf-8'),
                ##r[COL_CLUB].decode(u'utf-8')),
                ##inf])
        FMT = [(3, u'r'), u' ', (name_w, u'l'), (4, u'r')]
        self.meet.scbwin = scbwin.scbtable(scb=self.meet.scb,
                                           head=self.meet.racenamecat(
                                               self.event),
                                           subhead=u'STARTLIST',
                                           coldesc=FMT,
                                           rows=startlist)
        self.meet.scbwin.reset()

    def stat_but_cb(self, button):
        """Race ctrl button callback."""
        if self.timerstat in (u'idle', u'armstart'):
            self.armstart()
        elif self.timerstat in (u'running', u'armfinish'):
            self.armfinish()

    def checkplaces(self, places=u''):
        """Check the proposed places against current race model."""
        ret = True
        placeset = set()
        for no in strops.reformat_biblist(places).split():
            # repetition? - already in place set?
            if no in placeset:
                _log.error(u'Duplicate no in places: %r', no)
                ret = False
            placeset.add(no)
            # rider in the model?
            lr = self.getrider(no)
            if lr is None:
                if not self.meet.get_clubmode():
                    _log.error(u'Non-starter in places: %r', no)
                    ret = False
                # otherwise club mode allows non-starter in places
            else:
                # rider still in the race?
                if lr[COL_DNF]:
                    _log.error(u'DNF rider in places: %r', no)
                    ret = False
        return ret

    def race_ctrl_places_activate_cb(self, entry, data=None):
        """Respond to activate on place entry."""
        places = strops.reformat_placelist(entry.get_text().decode(u'utf-8'))
        if self.checkplaces(places):
            entry.set_text(places)
            self.do_places()
            glib.idle_add(self.delayed_announce)
            self.meet.delayed_export()
        else:
            _log.error(u'Places not updated')

    def race_ctrl_action_activate_cb(self, entry, data=None):
        """Perform current action on bibs listed."""
        rlist = entry.get_text().decode(u'utf-8')
        acode = self.action_model.get_value(
            self.ctrl_action_combo.get_active_iter(), 1).decode(u'utf-8')
        if acode == u'dnf':
            self.dnfriders(strops.reformat_biblist(rlist))
            entry.set_text('')
        elif acode == u'add':
            rlist = strops.riderlist_split(rlist, self.meet.rdb, self.series)
            for bib in rlist:
                self.addrider(bib)
            entry.set_text(u'')
        elif acode == u'del':
            rlist = strops.riderlist_split(rlist, self.meet.rdb, self.series)
            for bib in rlist:
                self.delrider(bib)
            entry.set_text(u'')
        elif acode == u'lap':
            self.runlap = strops.confopt_posint(rlist)
            _log.debug(u'Manually set runlap to: %r', self.runlap)
        elif acode == u'out' and self.evtype == u'elimination':
            bib = rlist.strip()
            if self.eliminate(bib):
                entry.set_text(u'')
            # Short-circuit method to avoid re-announce
            return False
        elif acode == u'in' and self.evtype == u'elimination':
            bib = rlist.strip()
            if self.uneliminate(bib):
                entry.set_text(u'')
            # Short-circuit method to avoid re-announce
            return False
        else:
            _log.error(u'Ignoring invalid action')
            return False
        glib.idle_add(self.delayed_announce)

    def update_expander_lbl_cb(self):
        """Update race info expander label."""
        self.info_expand.set_label(self.meet.infoline(self.event))

    def uneliminate(self, bib):
        """Remove rider from the set of eliminated riders."""
        ret = False
        r = self.getrider(bib)
        if r is not None:
            if not r[COL_DNF]:
                if bib in self.eliminated:
                    self.eliminated.remove(bib)
                    self.placexfer(
                        self.ctrl_places.get_text().decode(u'utf-8'))
                    _log.info(u'Rider %r removed from eliminated riders.', bib)
                    glib.idle_add(self.delayed_announce)
                    ret = True
                else:
                    _log.error(u'Rider %r not eliminated', bib)
            else:
                _log.error(u'Cannot un-eliminate dnf rider: %r', bib)
        else:
            _log.error(u'Cannot un-eliminate non-starter: %r', bib)

        return ret

    def eliminate(self, bib):
        """Register rider as eliminated."""
        ret = False
        r = self.getrider(bib)
        if r is not None:
            if not r[COL_DNF]:
                if bib not in self.eliminated:
                    self.eliminated.append(bib)
                    self.placexfer(
                        self.ctrl_places.get_text().decode(u'utf-8'))
                    _log.error(u'Rider %r eliminated.', bib)
                    ret = True
                    fname = r[COL_FIRSTNAME].decode(u'utf-8')
                    lname = r[COL_LASTNAME].decode(u'utf-8')
                    club = r[COL_CLUB].decode(u'utf-8')
                    rno = r[COL_BIB].decode(u'utf-8')
                    rstr = (rno + u' ' + strops.fitname(
                        fname, lname, self.meet.scb.linelen - 3 - len(rno)))
                    self.meet.scbwin = scbwin.scbintsprint(
                        scb=self.meet.scb,
                        line1=self.meet.racenamecat(self.event),
                        line2=u'RIDER ELIMINATED',
                        coldesc=[u' ', (self.meet.scb.linelen - 1, u'l')],
                        rows=[[rstr]])
                    self.meet.scbwin.reset()
                    self.meet.gemini.reset_fields()
                    self.meet.gemini.set_bib(bib)
                    self.meet.gemini.show_brt()
                    self.meet.cmd_announce(u'eliminated', bib)
                    # announce it:
                    nrstr = strops.truncpad(
                        strops.resname_bib(r[COL_BIB].decode(u'utf-8'),
                                           r[COL_FIRSTNAME].decode(u'utf-8'),
                                           r[COL_LASTNAME].decode(u'utf-8'),
                                           r[COL_CLUB].decode(u'utf-8')), 60)
                    self.meet.txt_postxt(21, 0, u'Out: ' + nrstr)
                    glib.timeout_add_seconds(10, self.delayed_result)
                else:
                    _log.error(u'Rider %r already eliminated.', bib)
            else:
                _log.error(u'Cannot eliminate dnf rider: %r', bib)
        else:
            _log.error(u'Cannot eliminate non-starter: %r', bib)

        return ret

    def delayed_result(self):
        if self.ctrl_action.get_property('has-focus'):
            if isinstance(self.meet.scbwin, scbwin.scbintsprint):
                FMT = [(3, u'l'), (3, u'r'), u' ',
                       (self.meet.scb.linelen - 11, u'l'), (4, u'r')]
                self.meet.scbwin = scbwin.scbtable(scb=self.meet.scb,
                                                   head=self.meet.racenamecat(
                                                       self.event),
                                                   subhead=self.resulttype,
                                                   coldesc=FMT,
                                                   rows=self.results)
                self.meet.scbwin.reset()
        self.meet.cmd_announce(u'eliminated', u'')
        self.meet.gemini.clear()
        glib.idle_add(self.delayed_announce)

    def editent_cb(self, entry, col):
        """Shared event entry update callback."""
        if col == u'pref':
            self.event[u'pref'] = entry.get_text().decode(u'utf-8')
        elif col == u'info':
            self.event[u'info'] = entry.get_text().decode(u'utf-8')
        self.update_expander_lbl_cb()

    def editcol_cb(self, cell, path, new_text, col):
        """Startlist cell update callback."""
        new_text = new_text.decode(u'utf-8').strip()
        if col == COL_BIB:
            if new_text.isalnum():
                if self.getrider(new_text) is None:
                    self.riders[path][COL_BIB] = new_text
                    dbr = self.meet.rdb.getrider(new_text, self.series)
                    if dbr is not None:
                        for i in range(1, 4):
                            self.riders[path][i] = self.meet.rdb.getvalue(
                                dbr, i)
        else:
            self.riders[path][col] = new_text.strip()

    def editcol_db(self, cell, path, new_text, col):
        """Cell update with writeback to meet."""
        new_text = new_text.decode(u'utf-8').strip()
        self.riders[path][col] = new_text
        glib.idle_add(self.meet.rider_edit,
                      self.riders[path][COL_BIB].decode(u'utf-8'), self.series,
                      col, new_text)

    def gotorow(self, i=None):
        """Select row for specified iterator."""
        if i is None:
            i = self.riders.get_iter_first()
        if i is not None:
            self.view.scroll_to_cell(self.riders.get_path(i))
            self.view.set_cursor_on_cell(self.riders.get_path(i))

    def dnf_cb(self, cell, path, col):
        """Toggle rider dnf flag."""
        self.riders[path][col] = not self.riders[path][col]

    def starttrig(self, e):
        """React to start trigger."""
        if self.timerstat == u'armstart':
            if self.distance and self.units == u'laps':
                self.runlap = self.distance - 1
                _log.debug(u'Set runlap: %r', self.runlap)
            self.start = e
            self.lstart = tod.now()
            self.setrunning()
            if self.timetype == u'200m':
                glib.timeout_add_seconds(4, self.armfinish)
                # delayed auto arm 200...

    def fintrig(self, e):
        """React to finish trigger."""
        if self.timerstat == u'armfinish':
            self.finish = e
            self.setfinished()
            self.set_elapsed()
            self.log_elapsed()
            if self.timerwin and type(self.meet.scbwin) is scbwin.scbtimer:
                self.showtimer()
            glib.idle_add(self.delayed_announce)

    def timercb(self, e):
        """Handle a timer event."""
        chan = timy.chan2id(e.chan)
        if chan == self.startchan or chan == 0:
            _log.debug(u'Got a start impulse')
            self.starttrig(e)
        elif chan == self.finchan:
            _log.debug('Got a finish impulse')
            self.fintrig(e)
        return False

    def timeout(self):
        """Update scoreboard and respond to timing events."""
        if not self.winopen:
            return False
        if self.finish is None:
            self.set_elapsed()
            if self.timerwin and type(self.meet.scbwin) is scbwin.scbtimer:
                self.meet.scbwin.settime(self.time_lbl.get_text())
        return True

    def race_info_time_edit_activate_cb(self, button):
        """Display race timing edit dialog."""
        ostx = u''
        oftx = u''
        if self.start is not None:
            ostx = self.start.rawtime(4)
        else:
            ostx = u'0.0'
        if self.finish is not None:
            oftx = self.finish.rawtime(4)
        ret = uiutil.edit_times_dlg(self.meet.window, ostx, oftx)
        if ret[0] == 1:
            try:
                stod = None
                if ret[1]:
                    stod = tod.tod(ret[1], u'MANU', u'C0i')
                    self.meet.main_timer.printline(u' ' + str(stod))
                ftod = None
                if ret[2]:
                    ftod = tod.tod(ret[2], u'MANU', u'C1i')
                    self.meet.main_timer.printline(u' ' + str(ftod))
                _log.info(u'Updating race times st=%r, ft=%r', stod, ftod)
                self.set_start(stod)
                self.set_finish(ftod)
                self.set_elapsed()
                if self.start is not None and self.finish is not None:
                    self.log_elapsed()
            except Exception as v:
                _log.error(u'Error updating times %s: %s',
                           v.__class__.__name__, v)

            glib.idle_add(self.delayed_announce)
        else:
            _log.info('Edit race times cancelled')

    def result_gen(self):
        """Generator function to export a final result."""
        ft = None
        for r in self.riders:
            bib = r[COL_BIB].decode(u'utf-8')
            rank = None
            info = u''
            if self.evtype in [u'handicap', u'sprint']:
                # include handicap and previous win info
                info = r[COL_INFO].decode(u'utf-8').strip()
            if self.onestart:
                if not r[COL_DNF]:
                    if r[COL_PLACE]:
                        rank = int(r[COL_PLACE].decode(u'utf-8'))
                else:
                    inft = r[COL_INFO].decode(u'utf-8')
                    if inft in [u'dns', u'dsq']:
                        rank = inft
                    else:
                        rank = u'dnf'
            time = None
            if self.finish is not None and ft is None:
                time = (self.finish - self.start).rawtime(2)
                ft = True
            yield [bib, rank, time, info]

    def result_report(self, recurse=False):
        """Return a list of report sections containing the race result."""
        self.placexfer(self.ctrl_places.get_text().decode(u'utf-8'))
        ret = []
        sec = report.section()
        sec.heading = u'Event ' + self.evno + u': ' + u' '.join(
            [self.event[u'pref'], self.event[u'info']]).strip()
        sec.lines = []
        lapstring = strops.lapstring(self.event[u'laps'])
        substr = u' '.join(
            [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
        first = True
        fs = u''
        if self.finish is not None:
            fs = self.time_lbl.get_text().decode(u'utf-8').strip()
        rcount = 0
        pcount = 0
        for r in self.riders:
            plstr = u''
            rcount += 1
            rno = r[COL_BIB].decode(u'utf-8')
            rh = self.meet.newgetrider(rno, self.series)
            rname = u''
            if rh is not None:
                rname = rh[u'namestr']
            inf = r[COL_INFO].decode(u'utf-8').strip()
            if self.evtype in [u'keirin', u'sprint']:  # encirc draw no
                inf = strops.drawno_encirc(inf)
            if r[COL_DNF]:
                pcount += 1
                if r[COL_INFO] in [u'dns', u'dsq']:
                    plstr = r[COL_INFO].decode(u'utf-8')
                    inf = None
                else:
                    plstr = u'dnf'
            elif self.onestart and r[COL_PLACE] != u'':
                plstr = r[COL_PLACE].decode(u'utf-8') + u'.'
                pcount += 1
            # but suppress inf if within an omnium
            if self.inomnium:
                inf = None
            if self.evtype != u'handicap' and rh is not None and rh[u'ucicode']:
                inf = rh[u'ucicode']  # overwrite by force
            if plstr:  # don't emit a row for unplaced riders
                if not first:
                    sec.lines.append([plstr, rno, rname, inf, None, None])
                else:
                    sec.lines.append([plstr, rno, rname, inf, fs, None])
                    first = False
        if self.onestart:
            substr = substr.strip()
            shvec = []
            if substr:
                shvec.append(substr)
            shvec.append(self.standingstr())
            sec.subheading = u' - '.join(shvec)
        else:
            if substr:
                sec.subheading = substr

        ret.append(sec)

        if len(self.comments) > 0:
            sec = report.bullet_text()
            sec.subheading = u'Decisions of the commisaires panel'
            for c in self.comments:
                sec.lines.append([None, c])
            ret.append(sec)
        return ret

    def standingstr(self, width=None):
        """Return an event status string for reports and scb."""
        ret = u''
        if self.onestart:
            ret = u'Standings'
            rcount = 0
            pcount = 0
            winner = False
            for r in self.riders:
                if r[COL_DNF]:
                    pcount += 1
                elif r[COL_PLACE] != u'':
                    if r[COL_PLACE] == u'1':
                        winner = True
                    pcount += 1
                rcount += 1
            if winner:
                if rcount > 0 and pcount < rcount:
                    ret = u'Provisional Result'
                else:
                    ret = u'Result'
        return ret

    def destroy(self):
        """Signal race shutdown."""
        self.frame.destroy()

    def show(self):
        """Show race window."""
        self.frame.show()

    def hide(self):
        """Hide race window."""
        self.frame.hide()

    def __init__(self, meet, event, ui=True):
        """Constructor.

        Parameters:

            meet -- handle to meet object
            event -- event object handle
            ui -- display user interface?

        """
        self.meet = meet
        self.event = event
        self.evno = event[u'evid']
        self.evtype = event[u'type']
        self.series = event[u'seri']
        self.configfile = meet.event_configfile(self.evno)
        self.results = []
        self.resulttype = u'RESULT'

        self.readonly = not ui
        rstr = u''
        if self.readonly:
            rstr = u'readonly '
        _log.debug(u'Init %sevent %s', rstr, self.evno)
        self.comments = []
        self.eliminated = []
        self.onestart = False
        self.runlap = None
        self.lastrunlap = None
        self.start = None
        self.lstart = None
        self.finish = None
        self.winopen = ui  # window 'open' on proper load- or consult edb
        self.timerwin = False
        self.timerstat = u'idle'
        self.distance = None
        self.units = u'laps'
        self.timetype = u'start/finish'
        self.startplace = 0  # offset to first place in this race (hack)
        self.autospec = u''  # automatic startlist
        self.inomnium = False
        self.seedsrc = None
        self.doscbplaces = True  # auto show result on scb
        self.reorderflag = 0
        self.startchan = 0
        self.finchan = 1

        self.riders = gtk.ListStore(
            gobject.TYPE_STRING,  # 0 bib
            gobject.TYPE_STRING,  # 1 first name
            gobject.TYPE_STRING,  # 2 last name
            gobject.TYPE_STRING,  # 3 club
            gobject.TYPE_STRING,  # 4 xtra info
            gobject.TYPE_BOOLEAN,  # 5 DNF/DNS
            gobject.TYPE_STRING)  # 6 placing

        uifile = os.path.join(metarace.UI_PATH, u'race.ui')
        _log.debug(u'Building event interface from %r', uifile)
        b = gtk.Builder()
        b.add_from_file(uifile)

        self.frame = b.get_object(u'race_vbox')
        self.frame.connect(u'destroy', self.shutdown)

        # info pane
        self.info_expand = b.get_object(u'info_expand')
        b.get_object(u'race_info_evno').set_text(self.evno)
        self.showev = b.get_object(u'race_info_evno_show')
        self.prefix_ent = b.get_object(u'race_info_prefix')
        self.prefix_ent.connect(u'changed', self.editent_cb, u'pref')
        self.prefix_ent.set_text(self.event[u'pref'])
        self.info_ent = b.get_object(u'race_info_title')
        self.info_ent.connect(u'changed', self.editent_cb, u'info')
        self.info_ent.set_text(self.event[u'info'])

        self.time_lbl = b.get_object(u'race_info_time')
        self.time_lbl.modify_font(uiutil.MONOFONT)

        # ctrl pane
        self.stat_but = uiutil.statbut(b.get_object(u'race_ctrl_stat_but'))
        self.ctrl_places = b.get_object(u'race_ctrl_places')
        self.ctrl_action_combo = b.get_object(u'race_ctrl_action_combo')
        self.ctrl_action = b.get_object(u'race_ctrl_action')
        self.action_model = b.get_object(u'race_action_model')

        # start timer and show window
        if ui:
            _log.debug(u'Connecting event ui handlers')
            # riders pane
            t = gtk.TreeView(self.riders)
            self.view = t
            t.set_reorderable(True)
            t.set_enable_search(False)
            t.set_rules_hint(True)

            # riders columns
            uiutil.mkviewcoltxt(t, u'No.', COL_BIB, calign=1.0)
            uiutil.mkviewcoltxt(t,
                                u'First Name',
                                COL_FIRSTNAME,
                                self.editcol_db,
                                expand=True)
            uiutil.mkviewcoltxt(t,
                                u'Last Name',
                                COL_LASTNAME,
                                self.editcol_db,
                                expand=True)
            uiutil.mkviewcoltxt(t, u'Club', COL_CLUB, self.editcol_db)
            uiutil.mkviewcoltxt(t, u'Info', COL_INFO, self.editcol_cb)
            uiutil.mkviewcolbool(t, u'DNF', COL_DNF, self.dnf_cb)
            uiutil.mkviewcoltxt(t,
                                u'Place',
                                COL_PLACE,
                                self.editcol_cb,
                                halign=0.5,
                                calign=0.5)
            t.show()
            b.get_object(u'race_result_win').add(t)
            b.connect_signals(self)
